#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from typing import List, Optional, Dict
from .csv import load_form_from_csv_file
from .excel import num_to_column, column_to_num
from .formatter_and_parser import is_none_or_empty, string_of, parse_value
from .error import Error

from .form import Form
from .utils import get_value_or_exception


class EasyForm:
    def __init__(self, err: Error, form: Form, must_has_content: bool = True):
        self.err = err
        self.__raw_form = form
        self.__all_fields: List[str] = list(map(lambda x: string_of(x), self.__raw_form.title_row.cells))
        self.___row_num_to_index_dict: Optional[Dict[int, int]] = self.__init_row_num_to_index_dict()
        self.__field_to_index_dict: Dict[str, int] = {}

        self.row_num_list: List[int] = list(map(lambda x: x.row_num, self.__raw_form.data_rows))
        if not self.err.has_error() and must_has_content and len(self.__raw_form.data_rows) == 0:
            msg = 'form is no data'
            self.err.append(msg)

    def __init_row_num_to_index_dict(self):
        if self.err.has_error():
            return
        result_dict: Dict[int, int] = {}

        index = 0
        for row in self.__raw_form.data_rows:
            result_dict[row.row_num] = index
            index += 1

        return result_dict

    def register_field(self, field: str, must_not_none=True):
        if self.err.has_error():
            return

        if field in self.__field_to_index_dict:
            return

        # get col index
        if field not in self.__all_fields:
            msg = 'cannot find field "{}" in title row'.format(field)
            self.err.append(msg)
            return

        index = self.__all_fields.index(field)

        # verify field data
        if must_not_none:
            for row in self.__raw_form.data_rows:
                val = row.cell(index, self.err)
                row_num = row.row_num
                if is_none_or_empty(val):
                    msg = 'empty cell found at row {}, field {}'.format(row_num, field)
                    self.err.append(msg)
                    return

        # register
        self.__field_to_index_dict[field] = index

    def register_multi_fields(self, fields: List[str], must_not_none=True):
        for field in fields:
            self.register_field(field, must_not_none)
            if self.err.has_error():
                return

    def __verify_field(self, field: str):
        if self.err.has_error():
            return

        if field not in self.__field_to_index_dict:
            msg = 'field "{}" is not register, or not exists in title row'.format(field)
            self.err.append(msg)

    def __verify_row_num(self, row_num: int):
        if self.err.has_error():
            return

        if row_num not in self.___row_num_to_index_dict:
            msg = 'row "{}" is out of range'.format(row_num)
            self.err.append(msg)

    def cell(self, field: str, row_num: int):
        self.__verify_field(field)
        self.__verify_row_num(row_num)
        if self.err.has_error():
            return

        field_index = self.__field_to_index_dict[field]
        row_index = self.___row_num_to_index_dict[row_num]

        return self.__raw_form.data_rows[row_index].cell(field_index, self.err)

    def excel_cell_range(self, field: str, row_num: int, start_column_name: str = 'A'):
        self.__verify_field(field)
        if self.err.has_error():
            return

        field_index = self.__field_to_index_dict[field]
        column_name = num_to_column(column_to_num(start_column_name) + field_index)
        return '{}{}'.format(column_name, row_num)

    def check_values_of_field(self, field: str, expected_value: any):
        self.__verify_field(field)
        if self.err.has_error():
            return

        row_count = len(self.__raw_form.data_rows)
        if row_count < 1:
            return

        field_index = self.__field_to_index_dict[field]
        for row_num in self.___row_num_to_index_dict:
            row_index = self.___row_num_to_index_dict[row_num]
            row_cell_value = self.__raw_form.data_rows[row_index].cell(field_index, self.err)

            if row_cell_value != expected_value:
                msg = 'column "{}" have unexpected value: {} at row {}, expected value: {}' \
                    .format(field, row_cell_value, row_num, expected_value)
                self.err.append(msg)
                return

    def format_cell_values(self, field: str, formatter=lambda x: x):
        self.__verify_field(field)
        if self.err.has_error():
            return

        field_index = self.__field_to_index_dict[field]
        for row in self.__raw_form.data_rows:
            val = row.cell(field_index, self.err)
            row.cells[field_index] = formatter(val)

    def find_row_num_list(self, field: str, val, data_type: str = "自动", must_have: bool = False,
                          must_only_one: bool = False, in_rows: Optional[List[int]] = None):
        self.__verify_field(field)
        if self.err.has_error():
            return

        field_index = self.__field_to_index_dict[field]
        row_num_list: List[int] = []
        for row_num in self.___row_num_to_index_dict:
            if in_rows is not None and row_num not in in_rows:
                continue
            row_index = self.___row_num_to_index_dict[row_num]
            row_cell_val = self.__raw_form.data_rows[row_index].cell(field_index, self.err)
            row_cell_val = get_value_or_exception(self.err, parse_value(row_cell_val, data_type=data_type))
            if self.err.has_error():
                return
            if val == row_cell_val:
                row_num_list.append(row_num)

        if self.err.has_error():
            return

        count = len(row_num_list)
        in_rows_desc = "" if in_rows is None else " , in rows: \'{}\'".format(",".join(in_rows))
        if count == 0 and must_have:
            msg = 'cannot find any row with value \'{}\' in field \'{}\'{}'.format(val, field, in_rows_desc)
            self.err.append(msg)
            return

        if count > 1 and must_only_one:
            msg = 'find too many rows \'{}\' with same value \'{}\' in field \'{}\'{}'\
                .format(row_num_list, val, field, in_rows_desc)
            self.err.append(msg)
            return

        return row_num_list

    def find_first_row_num(self, field: str, val, data_type: str = "自动", must_have: bool = True,
                           must_only_one: bool = True, in_rows: Optional[List[int]] = None):
        row_num_list = self.find_row_num_list(field, val, data_type=data_type, must_have=must_have,
                                              must_only_one=must_only_one,
                                              in_rows=in_rows)
        if self.err.has_error():
            return

        return row_num_list[0]


def build_easyform_from_csv_file(err: Error, csv_file_path: str):
    if err.has_error():
        return None

    form = load_form_from_csv_file(err, csv_file_path)
    if err.has_error():
        return None

    easy_form = EasyForm(err, form)
    if err.has_error():
        msg = 'last error occurred when read file {}'.format(csv_file_path)
        err.append(msg)
        return
    return easy_form



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# from datetime import datetime
#
# from grutils import utils
# from grutils import error
# from grutils.utils import date_of
# from grutils.excel import close_wb
# from grutils.form_utils import read_sht
# from grutils.utils import smart_repair_title


# print(utils.int_value_of("101.12", err, 0))
#
# print(date_of(datetime.date(datetime.now())))
# from grutils.error import Error
# from grutils.formatter_and_parser import parse_value
# from grutils.utils import get_value_or_exception
#
# e = Exception('aaa')
# _err = Error()
# v = parse_value('123.5', "日期")
# v = get_value_or_exception(_err, v, 'exception: {}')
# if _err.has_error():
#     print(_err.msg())
# else:
#     print(v)



# from grutils.progress import shared_progress
# from grutils.form_utils import read_sht
#
# shared_progress.reset()
#
# shared_progress.register_step("step 1", 6)
# shared_progress.register_step("step 2", 4)
# shared_progress.register_step("step 2.1", 0.6, ["step 2"])
# shared_progress.register_step("step 2.2", 0.4, ["step 2"])
# shared_progress.register_step("step 2.2.1", 1, ["step 2", "step 2.2"])
# shared_progress.register_step("step 2.2.2", 1, ["step 2", "step 2.2"])
# shared_progress.register_step("step 2.2.3", 1, ["step 2", "step 2.2"])
# shared_progress.register_step("step 2.2.4", 1, ["step 2", "step 2.2"])
# shared_progress.register_step("step 2.3", 1, ["step 2"])
# shared_progress.register_step("step 3", 10)
#
# shared_progress.re_assign_steps()
#
# shared_progress.finish_step(["step 3"])
# shared_progress.dump_steps()
#
# shared_progress.reset()
# shared_progress.finish_step(["step 2", "step 2.2", "step 2.2.1"])
# shared_progress.dump_steps()


# print('\n\n  ============== ')
# shared_progress.finish_step(["step 1"])
#
# print('\n\n  ============== ')
# shared_progress.finish_step(["step 2", "step 2.1"])
#
# print('\n\n  ============== ')
# shared_progress.finish_step(["step 2", "step 2.2", "step 2.2.1"])
# shared_progress.finish_step(["step 2", "step 2.2", "step 2.2.2"])
# shared_progress.finish_step(["step 2", "step 2.2", "step 2.2.3"])
# shared_progress.finish_step(["step 2", "step 2.2"])
#
# print('\n\n  ============== ')
# shared_progress.finish_step(["step 3"])
#
# print('\n\n  ============== ')
# shared_progress.finish()
# shared_progress.dump_steps()
#
# print('\n\n  ============== ')
# shared_progress.reset()
# shared_progress.dump_steps()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import os
import tkinter
from tkinter import W, filedialog, ttk, E
from typing import Optional, List

from .formatter_and_parser import is_none_or_empty


def __ask_path(title: str, old_path: str, filetypes: Optional[List[str]]):
    if filetypes is None:
        new_path = filedialog.askdirectory(
            initialdir=os.path.dirname(old_path),
            title=title
        )
    else:
        filetypes_name = " ".join(filetypes) + " files"
        filetypes_value = " ".join(list(map(lambda x: "." + x, filetypes)))
        new_path = filedialog.askopenfilename(
            initialdir=os.path.dirname(old_path),
            title=title,
            filetypes=[(filetypes_name, filetypes_value)]
        )

    if not is_none_or_empty(new_path) and new_path != '' and new_path != old_path:
        return new_path.replace('/', '\\')

    return None


def create_general_path_picker(frame: tkinter.Frame,
                               row: int,
                               old_path: str,
                               btn_title: str,
                               filetypes: Optional[List[str]],
                               on_changed: Optional[any] = None,
                               btn_width=30,
                               label_width=90,
                               pad=5
                               ):
    old_path = '' if old_path is None else old_path

    content = tkinter.StringVar(value=old_path)

    message = ttk.Entry(frame, width=label_width, text=content, style='TLabel', justify='right', state='readonly')
    message.grid(padx=pad, pady=pad, row=row, column=1, columnspan=3, sticky=E + W)

    def on_btn_click():
        new_path = __ask_path(btn_title, old_path, filetypes)
        if new_path is not None:
            content.set(new_path)
            if on_changed is not None:
                on_changed(new_path, old_path)

    btn = tkinter.Button(frame, width=btn_width, text=btn_title, command=on_btn_click)
    btn.grid(padx=pad, pady=pad, row=row, column=0, columnspan=1, sticky=E + W)

    return btn, message



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from typing import Dict, Optional


class GUIEvent:
    def __init__(self, category: str, name: str, key: Optional[str] = None, val: any = None):
        self.category = category
        self.name = name
        self.payload: Dict[str, any] = {}
        if key is not None and val is not None:
            self.payload[key] = val

    def add_property(self, k: str, v: any):
        self.payload[k] = v
        return self

    def __repr__(self):
        s = "\nGUI Event<\n"
        s += "category: {}\n".format(self.category)
        s += "name: {}\n".format(self.name)
        s += "payload: {}\n".format(self.payload)
        s += ">\n"
        return s

    def __str__(self):
        return self.__repr__()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import os
from typing import Dict, List, Optional, Tuple

from .configs import Configuration
from .error import Error


def sort_and_merge_ints(ints: List[int]):
    results: List[Tuple[int, int]] = []
    if len(ints) == 0:
        return results

    sorted_ints = sorted(ints)

    start: Optional[int] = None
    curr: Optional[int] = None
    for i in sorted_ints:
        if start is None:
            start = i
            curr = i
            continue

        if i == curr + 1:
            curr = i
            continue

        results.append((start, curr))
        start = i
        curr = i

    if start is not None:
        results.append((start, curr))

    return results


def get_simple_desc_of_int_groups(groups: List[Tuple[int, int]],
                                  formatter: str = "{}",
                                  linker: str = "-",
                                  joiner: str = ","
                                  ):
    items = list(map(lambda group: (formatter + linker + formatter).format(group[0], group[1]) if group[0] != group[
        1] else formatter.format(group[0]), groups))
    return joiner.join(items)


def compare_two_list(a: Optional[List], b: Optional[List]):
    if a is None and b is None:
        return True

    if a is None or b is None:
        return False

    for k in a:
        if k not in b:
            return False

    for k in b:
        if k not in a:
            return False

    return True


def get_os_user():
    return os.getenv('username')


def get_special_folder(err: Error, key: str, config: Configuration, test_os_user: str = 'liuleidong'):
    if err.has_error():
        return

    folder = config.get(key=key, default='', save_if_no=False)
    if os.path.exists(folder):
        return folder

    os_user = get_os_user()
    if os_user == test_os_user:
        folder = config.get(key='test_' + key, default='', save_if_no=False)
        if len(folder) == 0:
            return folder
        if os.path.exists(folder):
            return folder

    err.append('folder "' + folder + '" is not exists')


def get_from_dict(d: Dict, *keys):
    _d = d
    for key in keys:
        if key in _d:
            _d = _d[key]
        else:
            return None

    return _d


def set_to_dict(d: Dict, v, *keys):
    _d = d
    for key in keys[:-1]:
        if key not in _d:
            _d[key] = {}
        _d = _d[key]

    _d[keys[-1]] = v


def get_value_or_exception(err: Error, v_or_e, exception_formatter='{}'):
    if err.has_error():
        return

    if type(v_or_e) == Exception:
        err.append(exception_formatter.format(v_or_e))
        return
    return v_or_e


def smart_repair_title(title: str):
    if type(title) != str:
        return title
    parts = title.split(" ")
    parts = list(filter(lambda x: x != "" and x != " ", parts))
    return " ".join(parts)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from pprint import pprint
from typing import List, Optional, Dict


class Progress:
    def __init__(self):
        self.curr = 0
        self.max = 10000
        self.progress_bar = None
        self.__root_step_data: Dict = {
            'percent': 1.0,
            'is_finished': False
        }

    def set_process_bar(self, process_bar):
        self.progress_bar = process_bar
        if self.progress_bar is not None:
            self.progress_bar["value"] = self.curr
            self.progress_bar["maximum"] = self.max

    def set(self, val: int):
        if val < 0:
            val = 0

        self.curr = val
        print("[progress] percent:", self.curr_percent())
        if self.progress_bar is not None:
            self.progress_bar["value"] = min(self.curr, self.max)
            self.progress_bar.update()

    def set_percent(self, percent: float):
        self.set(int(percent * self.max))

    def curr_percent(self):
        return float(self.curr) / float(self.max)

    def increase_percent(self, delta_percent: float):
        percent = self.curr_percent() + delta_percent
        self.set_percent(percent)

    def reset(self, unregister_all_steps: bool = False):
        if unregister_all_steps:
            self.__root_step_data = {
                'percent': 1.0,
                'is_finished': False
            }
        else:
            self.__set_step_status(False)
        self.set(0)

    def finish(self):
        self.finish_step()

    def remaining(self):
        return self.max - self.curr

    def remaining_percent(self):
        return 1.0 - self.curr_percent()

    def __find_step(self, chain: Optional[List[str]] = None):
        step_chain: List[Dict] = [self.__root_step_data]

        if chain is None:
            return step_chain

        for step_name in chain:
            father_step = step_chain[-1]
            if 'steps' not in father_step or step_name not in father_step['steps']:
                print('[progress] warning: step {} in chain {} not found'.format(step_name, chain))
                return None

            step_chain.append(father_step['steps'][step_name])

        return step_chain

    def register_step(self, step: str, weight: float, chain: Optional[List[str]] = None):
        step_chain = self.__find_step(chain)
        if step_chain is None:
            return

        father_step = step_chain[-1]
        if 'steps' not in father_step:
            father_step['steps'] = {}
        father_steps = father_step['steps']

        father_steps[step] = {
            'percent': weight,
            'is_finished': False
        }

    def re_assign_steps(self):
        def re_assign_sub_steps(d: Dict):
            if 'steps' not in d:
                return

            d_subs_total_percent = 0.0
            for step in d['steps']:
                d_subs_total_percent += d['steps'][step]['percent']

            if d_subs_total_percent == 0.0:
                return

            for step in d['steps']:
                d_sub = d['steps'][step]
                re_assign_sub_steps(d_sub)
                d_sub['percent'] = round((d_sub['percent'] / d_subs_total_percent) * 100000) / 100000

        re_assign_sub_steps(self.__root_step_data)

    def __set_step_status(self, is_finished: bool, chain: Optional[List[str]] = None):
        def reset_step(d: Dict):
            d['is_finished'] = is_finished
            if 'steps' in d:
                for step in d['steps']:
                    reset_step(d['steps'][step])

        step_chain = self.__find_step(chain)
        if step_chain is not None:
            reset_step(step_chain[-1])
        if is_finished:
            _chain = ["@"] if chain is None else ["@"] + chain
            print('[progress] step: {} finished'.format("=>".join(_chain)))

    def __cal_percent_from_step_data(self):
        def percent_of(d: Dict, d_father_total_percent: float):
            d_total_percent = d_father_total_percent * d['percent']
            if d['is_finished']:
                return d_total_percent

            if 'steps' not in d:
                return 0.0

            d_subs_curr_percent = 0.0
            for step in d['steps']:
                d_subs_curr_percent += percent_of(d['steps'][step], d_total_percent)
            return d_subs_curr_percent

        return percent_of(self.__root_step_data, 1.0)

    def finish_step(self, chain: Optional[List[str]] = None):
        self.__set_step_status(True, chain)

        self.set_percent(self.__cal_percent_from_step_data())

    def dump_steps(self):
        pprint(self.__root_step_data)


shared_progress = Progress()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import tkinter
from tkinter import W, END, ttk, E


def create_log_view(frame: tkinter.Frame, row: int, initial_content: str = '', height=30):
    row += 1
    logs_view = tkinter.Text(frame, height=height)
    logs_view.grid(padx=5, pady=5, row=row, column=0, columnspan=4, sticky=E + W)
    logs_view_scrollbar = ttk.Scrollbar(frame, command=logs_view.yview)
    logs_view_scrollbar.grid(row=row, column=6, sticky='nsew')
    logs_view['yscrollcommand'] = logs_view_scrollbar.set
    append_log(logs_view, initial_content)
    return logs_view


def append_log(logs_view: tkinter.Text, log: str, is_err: bool = False):
    if logs_view is None:
        return

    logs_view.insert(END, log + '\n')

    # auto scroll to bottom
    logs_view.see(END)
    if is_err:
        set_logs_color(logs_view, 'red')


def clear_logs(logs_view: tkinter.Text):
    if logs_view is None:
        return
    logs_view.delete(1.0, END)
    set_logs_color(logs_view, 'white')


def set_logs_color(logs_view: tkinter.Text, color: str = 'white'):
    if logs_view is None:
        return
    logs_view.config(background=color)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from typing import List
from .excel import row_items, num_to_column, column_to_num, column_items
from .formatter_and_parser import is_none_or_empty
from .error import Error
import xlwings as xw
from .form import Form


def remove_unfilled_rows(err: Error, form: Form, must_not_empty_fields: List[str]):
    if err.has_error():
        return

    field_indexes: List[int] = list(map(lambda x: form.column_index_with_title(x), must_not_empty_fields))
    if err.has_error():
        return

    unfilled_row_indexes: List[int] = []
    for row_index in range(len(form.data_rows)):
        row = form.data_rows[row_index]
        for field_index in field_indexes:
            cell_val = row.cell(field_index, err)
            if is_none_or_empty(cell_val):
                unfilled_row_indexes.insert(0, row_index)
                break

    for row_index in unfilled_row_indexes:
        form.data_rows.pop(row_index)


def read_sht(err: Error, sht: xw.Sheet, ref_col: str = 'A', header_row: int = 1, start_col: str = 'A'):
    cells_row = row_items(err, sht, header_row, start_col)
    if err.has_error():
        return

    if len(cells_row) == 0:
        return
    end_col = num_to_column(column_to_num(start_col) + len(cells_row) - 1)

    form = Form(err)
    form.set_title_row(header_row, cells_row)
    cells_col = column_items(err, sht, ref_col, header_row + 1)
    for i in range(len(cells_col)):
        row_num = header_row + 1 + i
        cells = sht.range("{}{}:{}{}".format(start_col, row_num, end_col, row_num)).value
        form.append_data_row(row_num, cells)

    return form



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from datetime import datetime, date
from typing import List


def is_none_or_empty(s: any):
    return (type(s) == str and len(s) == 0) or (s is None)


def strip_if_str(s):
    if type(s) == str:
        return s.strip()
    else:
        return s


def parse_as_str_if_float_or_int(v):
    if type(v) == float:
        i = int(round(v, 0))
        return '{}'.format(i)
    elif type(v) == int:
        return '{}'.format(v)
    else:
        return v


def date_of(a: any):
    t = type(a)
    if t == str:
        if a.find('/') >= 0:
            items: List[str] = a.split('/')
            items_count = len(items)
            if items_count == 2:
                month = int(items[0])
                day = int(items[1])
                today = datetime.date(datetime.now())
                month_diff = today.month - month
                year = today.year if (0 <= month_diff <= 8) else (today.year + 1)
                return date.fromisoformat('{:04d}-{:02d}-{:02d}'.format(year, month, day))
            elif items_count == 3:
                month = int(items[0])
                day = int(items[1])
                year = int(items[2])
                return date.fromisoformat('{:04d}-{:02d}-{:02d}'.format(year, month, day))
        elif a.find('-') >= 0:
            items: List[str] = a.split('-')
            if len(items) == 3:
                year = int(items[0])
                month = int(items[1])
                day = int(items[2])
                return date.fromisoformat('{:04d}-{:02d}-{:02d}'.format(year, month, day))
        raise Exception('not found splitter / or - in raw string')
    elif t == datetime:
        return a.date()
    elif t == date:
        return a
    else:
        raise Exception('invalid raw data type: {}'.format(t))


def float_02(f: float):
    return round(f * 100) / 100


def str_without_prefix(val: any, prefix='\''):
    if type(val) == str and val.startswith(prefix):
        return val[len(prefix):]
    else:
        return val


def str_without_prefixes(val: any, prefixes: List[str]):
    if type(val) != str:
        return val
    _val = val
    for prefix in prefixes:
        if _val.startswith(prefix):
            _val = _val[len(prefix):]
    return _val


def str_with_prefix(val: any, formatter: str = '{}', prefix='\''):
    if type(val) == str and val.startswith(prefix):
        return val
    else:
        return (prefix + formatter).format(val)


def simple_date_str(d: date):
    return '{}/{}'.format(d.month, d.day)


def string_of(v: any):
    return v if type(v) == str else '{}'.format(v)


def bool_of(v: any):
    if type(v) == str:
        _v = v.lower().strip()
        return _v in ["是", "true", "t"]
    elif type(v) == bool:
        return v
    else:
        raise Exception('invalid type {}'.format(type(v)))


# noinspection PyBroadException
def parse_value(val, data_type: str = "自动", raise_exception=True):
    if data_type == "自动":
        return val

    if isinstance(val, list):
        return list(map(lambda x: parse_value(x, data_type, raise_exception), val))

    _val = strip_if_str(val)
    try:
        if data_type == "文本":
            _val = parse_as_str_if_float_or_int(val)
            if type(_val) != str:
                _val = '{}'.format(_val)
            return _val

        if type(_val) == str:
            _val = str_without_prefixes(_val, ["'", "`"])
        if data_type == "整数":
            return int(round(float(_val)))

        if data_type == "小数":
            return float(_val)

        if data_type == "日期":
            return date_of(_val)

        if data_type == "是否":
            return bool_of(_val)
    except Exception as e:
        if not raise_exception:
            return None

        msg = 'failed with parse {} from ({}), details: {}'.format(data_type, val, e)
        return Exception(msg)

    msg = 'unsupported data type ({})'.format(data_type)
    return Exception(msg)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import os
import shutil
from typing import List, Union, Callable
import xlwings as xw
from xlwings.utils import rgb_to_int
from .error import Error
from .form import Form


def sheet_with_name(err: Error, wb: xw.Book, name='sheet name'):
    if err.has_error():
        return None
    for sht in wb.sheets:
        if sht.name == name:
            return sht

    err.append('excel file({}) has no sheet named "{}"'.format(wb.fullname, name))
    return None


def count_of_continue_none_cells(items: List[any] = None):
    if items is None:
        items = []
    i = len(items) - 1
    count = 0
    while i >= 0:
        if items[i] is None:
            count += 1
        else:
            break
        i -= 1
    return count


def row_items(err: Error, sht: xw.Sheet, row=1, first_column='A', last_column='IV'):
    if err.has_error():
        return None
    range_str = '{}{}:{}{}'.format(first_column, row, last_column, row)
    r = sht.range(range_str)
    cells = r.value

    count = count_of_continue_none_cells(cells)
    return cells[0:len(cells) - count]


def row_items_with_column(err: Error, sht: xw.Sheet, row=1, first_column='A', last_column='IV'):
    items = row_items(err, sht, row, first_column, last_column)
    results = []
    i = 0
    for item in items:
        results.append((add_column(first_column, i), item))
        i += 1

    return results


def row_items_filtered(err: Error, sht: xw.Sheet, tester=lambda x: True, row=1, first_column='A', last_column='IV'):
    items = row_items_with_column(err, sht, row, first_column, last_column)
    if tester is None:
        return items

    results = filter(lambda x: tester(x[1]), items)
    return list(results)


def column_items_sub(err: Error, sht: xw.Sheet, column='A', start_row=1, steps=100):
    if err.has_error():
        return []
    end_row = start_row + steps - 1
    if type(column) == str:
        range_str = '{}{}:{}{}'.format(column, start_row, column, end_row)
        return sht.range(range_str).value
    else:
        return sht.range((start_row, column), (end_row, column)).value


def column_items(err: Error, sht: xw.Sheet, column='A', start_row=1, steps=100):
    if err.has_error():
        return []

    items = []
    start_row_sub = start_row
    count = count_of_continue_none_cells(items)
    while count < 20:
        items = items + column_items_sub(err, sht, column, start_row_sub, steps)
        start_row_sub += steps
        count = count_of_continue_none_cells(items)

    return items[0:len(items) - count]


def column_items_with_row(err: Error, sht: xw.Sheet, column='A', start_row=1, steps=100):
    items = column_items(err, sht, column, start_row, steps)
    results = []
    i = 0
    for item in items:
        results.append((start_row + i, item))
        i += 1

    return results


def column_items_filtered(err: Error, sht: xw.Sheet, tester=lambda x: True, column='A', start_row=1, steps=100):
    items = column_items_with_row(err, sht, column, start_row, steps)
    if tester is None:
        return items

    results = filter(lambda x: tester(x[1]), items)
    return list(results)


def num_to_column(num: int):
    if num < 0:
        s = 'column number({}) is less than 0'.format(num)
        raise Exception(s)
    elif num == 0:
        return ""
    elif num <= 26:
        return chr(65 - 1 + num)

    right = num % 26
    if right == 0:
        right = 26
    left = int((num - right) / 26)
    return num_to_column(left) + num_to_column(right)


def column_to_num(column: str):
    length = len(column)
    if length == 0:
        return 0
    elif length == 1:
        num = ord(column.upper()) - ord('A') + 1
        if num < 1 or num > 26:
            s = 'column char({}) is not in a~z, or A~Z'.format(column)
            raise Exception(s)
        return num
    else:
        return 26 * column_to_num(column[0:length - 1]) + column_to_num(column[length - 1:])


def add_column(column: str, add_num: int):
    num = column_to_num(column)
    return num_to_column(num + add_num)


def close_wb(wb: xw.Book):
    if wb is not None:
        wb.close()


def quit_app(app: xw.App):
    if app is not None:
        app.quit()


def orange_red():
    return rgb_to_int((255, 69, 0))


def banana_yellow():
    return rgb_to_int((227, 207, 87))


def sky_blue():
    return rgb_to_int((135, 206, 235))


def green():
    return rgb_to_int((0, 255, 0))


def is_excel_file_opened(file_path: str):
    (file_folder, file_name) = os.path.split(file_path)

    locked_file = file_folder + '\\~$' + file_name
    return os.path.exists(locked_file)


def append_sht_to_another(err: Error, source_wb: xw.Book, target_wb: xw.Book,
                          source_sheet: str = 'Sheet1',
                          target_sheet: str = 'Sheet1',
                          skip_if_no_source_sheet: bool = False,
                          empty_rows: int = 1,
                          source_ref_column: str = 'A',
                          target_ref_column: str = 'A',
                          source_start_row: int = 1,
                          target_start_row: int = 1):
    if err.has_error():
        return
    source_sht = sheet_with_name(err, source_wb, source_sheet)
    if err.has_error():
        if skip_if_no_source_sheet:
            print('[Info] skip to copy sheet "{}", because it is not exists'.format(source_sheet))
            err.clear()
            return
        return

    source_row_count = len(column_items(err, source_sht, source_ref_column, start_row=source_start_row))
    if err.has_error():
        return

    if source_row_count == 0:
        return

    target_sht = sheet_with_name(err, target_wb, target_sheet)
    if err.has_error():
        return
    target_row_count = len(column_items(err, target_sht, target_ref_column, start_row=target_start_row))
    if err.has_error():
        return

    source_range = source_sht.range('A{}:IV{}'.format(source_start_row, source_start_row + source_row_count - 1))
    split_row_count = 0 if target_row_count == 0 else empty_rows

    _target_start_row = target_start_row - 1 + target_row_count + 1 + split_row_count
    _target_end_row = target_start_row - 1 + target_row_count + source_row_count + split_row_count
    target_range = target_sht.range('A{}:IV{}'.format(_target_start_row, _target_end_row))
    source_range.copy(target_range)
    return True


def upload_data_to_another_file(err: Error,
                                app: xw.App,
                                source_file_path: str,
                                target_file_path: str,
                                source_sheet: str = 'Sheet1',
                                target_sheet: str = 'Sheet1',
                                empty_rows: int = 1,
                                source_ref_column: str = 'A',
                                target_ref_column: str = 'A'
                                ):
    if err.has_error():
        return

    # check target folder is exists
    target_folder = os.path.dirname(target_file_path)
    if not os.path.exists(target_folder):
        err.append('try to upload data to shared file {}, but its folder {} is not exists'
                   .format(target_file_path, target_folder))
        return

    # check file opening
    opened = is_excel_file_opened(target_file_path)
    if opened:
        err.append('try to upload data to shared file {}, but it is opened now'.format(target_file_path))
        return

    # copy file if target is not exists
    if not os.path.exists(target_file_path):
        try:
            shutil.copy(source_file_path, target_file_path)
            target_wb: xw.Book = app.books.open(target_file_path)

            # rename sheet if should
            if source_sheet != target_sheet:
                target_sht = sheet_with_name(err, target_wb, source_sheet)
                if err.has_error():
                    close_wb(target_wb)
                    return
                target_sht.name = target_sheet
                target_wb.save()

            # remove useless sheets
            has_useless_sheets = False
            for sht in target_wb.sheets:
                if sht.name != target_sheet:
                    has_useless_sheets = True
                    sht.delete()
            if has_useless_sheets:
                target_wb.save()

            close_wb(target_wb)

        except Exception as e:
            err.append(
                'copy file {} to file {} failed with error {}'.format(source_file_path, target_file_path, e))
        return

    # uploading
    source_wb = app.books.open(source_file_path)
    target_wb = app.books.open(target_file_path)

    uploaded = append_sht_to_another(err, source_wb, target_wb,
                                     source_sheet,
                                     target_sheet,
                                     False,
                                     empty_rows,
                                     source_ref_column,
                                     target_ref_column)

    if uploaded is not None and uploaded:
        target_wb.save()

    close_wb(source_wb)
    close_wb(target_wb)


def get_blocks_from_sheet(err: Error, sht: xw.Sheet,
                          block_mark_tester: Union[str, Callable[[any], bool]],
                          block_mark_col='A',
                          block_title_row_ref_num=0,
                          block_start_col='A',
                          start_row=1
                          ):
    if err.has_error():
        return

    # find block mark rows
    items = column_items_with_row(err, sht, column=block_mark_col, start_row=start_row)
    if err.has_error():
        return

    def __tester(x):
        match = block_mark_tester(x) if type(block_mark_tester) != str else '{}'.format(x) == block_mark_tester
        return match

    mark_column_cells = list(filter(lambda x: __tester(x[1]), items))
    blocks_count = len(mark_column_cells)
    if blocks_count == 0:
        msg = 'cannot find any block marked in column \'{}\' with start row \'{}\''.format(block_mark_col, start_row)
        err.append(msg)
        return

    # read block one by one
    last_row_num = items[-1][0]
    blocks: List[Form] = []
    for i in range(blocks_count):
        mark_cell = mark_column_cells[i]
        block_mark_row_num, mark_value = mark_cell
        block_title_row_num = block_mark_row_num + block_title_row_ref_num
        form = Form(err)

        # check block content rows count
        block_end_row_num = last_row_num if i == blocks_count - 1 else mark_column_cells[i + 1][0] - 1
        block_row_count = block_end_row_num - block_title_row_num + 1
        if block_row_count < 2:
            msg = 'block at row \'{}\' is empty or only contains title row' \
                .format(block_mark_row_num)
            err.append(msg)
            return

        # read block title
        block_title_cells = row_items(err, sht, block_title_row_num, block_start_col)
        if err.has_error():
            return

        if len(block_title_cells) == 0:
            msg = 'title row of block at row \'{}\' is empty' \
                .format(block_mark_row_num)
            err.append(msg)
            return

        form.set_title_row(block_title_row_num, block_title_cells)

        block_end_col = num_to_column(column_to_num(block_start_col) + len(block_title_cells) - 1)

        # read block content
        for j in range(block_row_count - 1):
            row_num = block_title_row_num + 1 + j
            phase_content_row_range_str = "{}{}:{}{}".format(block_start_col, row_num, block_end_col, row_num)
            cells = sht.range(phase_content_row_range_str).value

            form.append_data_row(row_num, cells)

        # save form
        if err.has_error():
            return

        blocks.append(form)

    return None if err.has_error() else blocks


def build_row_range(row_num: int):
    return 'A{}:IV{}'.format(row_num, row_num)


def copy_excel_row(source_sht: xw.Sheet, source_row_num: int,
                   target_sht: xw.Sheet, target_row_num: int,
                   with_content: bool = False):
    source_range = build_row_range(source_row_num)
    target_range = build_row_range(target_row_num)
    source_sht.range(source_range).copy(target_sht.range(target_range))
    if not with_content:
        target_sht.range(target_range).clear_contents()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from typing import Dict, List, Optional


def build_warning(file: str, row: any, details: str):
    return build_warning_v2(file, "", row, details)


def build_warning_v2(file: str, sheet: str, row: any, details: str):
    if hasattr(row, "__iter__") and len(row) == 1:
        row_ = '{}'.format(row[0])
    else:
        row_ = '{}'.format(row)
    return {
        'file': file,
        'sheet': sheet,
        'row': row_,
        'details': details
    }


class Error:
    def __init__(self):
        self.msgs = []
        self.warnings = []

    def has_error(self):
        return len(self.msgs) > 0

    def append(self, msg='undefined error'):
        self.msgs.append('[Error] ' + msg)

    def clear(self):
        self.msgs = []

    def msg(self):
        return '; '.join(self.msgs)

    def has_warning(self):
        return len(self.warnings) > 0

    def append_warning(self, warning: Dict[str, str]):
        self.warnings.append(warning)

    def append_warnings(self, warnings: List[Dict[str, str]]):
        if len(warnings) > 0:
            for warning in warnings:
                self.warnings.append(warning)

    def clear_warnings(self):
        self.warnings = []

    def add_err_details(self, details: any, details_key: any = None):
        if len(self.msgs) == 0:
            return

        if details_key is None:
            self.msgs[-1] = '{} <{}>'.format(self.msgs[-1], details)
        else:
            self.msgs[-1] = '{} <{}: {}>'.format(self.msgs[-1], details_key, details)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from typing import List

from .error import Error
from .formatter_and_parser import strip_if_str
from .iterable import first_match
from .utils import smart_repair_title


class Row:
    def __init__(self, row_num=-1, cells: List[any] = None, do_strip=True):
        if cells is None:
            cells = []
        self.row_num = row_num
        if do_strip:
            self.cells = list(map(lambda x: strip_if_str(x), cells))
        else:
            self.cells = cells

    def __repr__(self):
        return '\nRow<row_num: {}, cells: {}>' \
            .format(self.row_num, self.cells)

    def __str__(self):
        return self.__repr__()

    def is_inited(self):
        return (self.row_num >= 0) and (self.cells is not None)

    def index(self, cell, err=None):
        i = first_match(lambda x: x == cell, self.cells).index
        if err is not None and i < 0:
            err.append('can not find cell ({})'.format(cell))
        return i

    def cell(self, column_index: int, err: Error):
        if not self.is_inited():
            err.append('row is not inited')
            return None
        if column_index >= len(self.cells) or column_index < 0:
            err.append(
                'column index({}) is out of range of row({})\' cells\' count {}'.format(column_index, self.row_num,
                                                                                        len(self.cells)))
            return None

        return self.cells[column_index]

    def has_content(self):
        if not self.is_inited():
            return False

        for cell in self.cells:
            if len(cell) > 0:
                return True

        return False

    def string(self, splitter: str = ',', line_ending: str = '\r\n'):
        return splitter.join(self.cells) + line_ending


class Form:
    def __init__(self, err: Error):
        self.err = err
        self.title_row = Row(-1, [])
        self.data_rows: List[Row] = []

    def __repr__(self):
        return '\nForm<title: {}, data: {}>' \
            .format(self.title_row, self.data_rows)

    def __str__(self):
        return self.__repr__()

    def row_with_num(self, row_num=0):
        return first_match(lambda x: x.row_num == row_num, self.data_rows).data

    def column_index_with_title(self, title=''):
        column_index = self.title_row.index(title)
        if column_index < 0:
            self.err.append('can not find title "{}"'.format(title))
        return column_index

    def title_to_column_index_mapping(self, titles: List[str]):
        if self.err.has_error():
            return

        mapping = {}
        for title in titles:
            index = self.column_index_with_title(title)
            if self.err.has_error():
                return
            mapping[title] = index

        return mapping

    def set_title_row(self, row_num=1, titles: List[any] = None, do_strip=True):
        if titles is None:
            titles = []
        titles = list(map(lambda x: smart_repair_title(x), titles))
        r = self.title_row
        if r.is_inited():
            self.err.append('title row has already been included')
        else:
            self.title_row = Row(row_num, titles, do_strip)

    def append_data_row(self, row_num=2, cells: List[any] = None, do_strip=True):
        if cells is None:
            cells = []
        if self.row_with_num(row_num) is not None:
            self.err.append('row with number {} has already been included'.format(row_num))
        else:
            if not self.title_row.is_inited():
                self.err.append('should include title row before append data')
                return
            diff = len(self.title_row.cells) - len(cells)
            while diff > 0:
                cells.append(None)
                diff -= 1
            self.data_rows.append(Row(row_num, cells, do_strip))

    def row(self, row_index=0):
        if self.err.has_error():
            return None
        if row_index >= len(self.data_rows) or row_index < 0:
            self.err.append('row index {} is out of range'.format(row_index))
            return None
        return self.data_rows[row_index]

    def cell(self, row_index=0, column_index=0):
        r = self.row(row_index)
        if r is None:
            return None

        return r.cell(column_index, self.err)

    def cell_with_title(self, row_index=0, title=''):
        if self.err.has_error():
            return
        column_index = self.column_index_with_title(title)
        return self.cell(row_index, column_index)

    def cell_with_title_must_have(self, row_index=0, title=''):
        if self.err.has_error():
            return

        cell = self.cell_with_title(row_index, title)
        if self.err.has_error():
            return

        if cell is None:
            self.err.append('empty cell at row {} with column name {}'.format(self.data_rows[row_index].row_num, title))
            return

        return cell

    def find_row_by_column(self, column_title: str, column_value: any):
        column_index = self.title_row.index(column_title)
        if column_index < 0:
            self.err.append('column with title({}) is not found'.format(column_title))
            return None

        r = first_match(lambda x: x.cells[column_index] == column_value, self.data_rows)
        if r.data is None:
            self.err.append('cannot find row with column value({}) in column (title: {})'
                            .format(column_value, column_title))

        return r

    def row_with_value(self, column_title: str, column_value: any):
        r = self.find_row_by_column(column_title, column_value)
        return None if self.err.has_error() else r.data

    def row_index_with_value(self, column_title: str, column_value: any):
        r = self.find_row_by_column(column_title, column_value)
        return -1 if self.err.has_error() else r.index

    def cells_lookup(self, title, value_of_title, other_titles: List[any] = None):
        if other_titles is None:
            other_titles = []
        if self.err.has_error():
            return None
        r: Row = self.row_with_value(title, value_of_title)

        if self.err.has_error():
            return None

        results = []
        for t in other_titles:
            i = self.title_row.index(t)
            if i < 0:
                self.err.append('title({}) is not found'.format(t))
                return None
            else:
                results.append(r.cell(i, self.err))
                if self.err.has_error():
                    return None
        return results

    def cell_lookup(self, title, value_of_title, other_title=''):
        cells = self.cells_lookup(title, value_of_title, [other_title])
        if self.err.has_error() or cells is None or len(cells) == 0:
            return None

        return cells[0]

    def data_row_nums_with_content(self):
        nums = []
        for r in self.data_rows:
            if r.has_content():
                nums.append(r.row_num)
        return nums

    def export_as_csv_file(self, file_path: str, splitter: str = ',', line_ending: str = '\r\n'):
        with open(file_path, "w") as f:
            f.write(self.title_row.string(splitter, line_ending))

            for row in self.data_rows:
                f.write(row.string(splitter, line_ending))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from shutil import copy
from typing import Dict, Optional
from .error import Error
from .excel import is_excel_file_opened, sheet_with_name, close_wb, quit_app
from .file import should_exists
import xlwings as xw
from .formatter_and_parser import is_none_or_empty


class ExcelFilesManager:
    def __init__(self, visible=True, add_book=False):
        self.excel_app = xw.App(visible=visible, add_book=add_book)
        self.opened_excel_files: Dict[str, xw.Book] = {}

    def close_opened_excel_file(self, file_path: str, save_changes: bool = True):
        if file_path in self.opened_excel_files:
            wb = self.opened_excel_files[file_path]
            if save_changes:
                wb.save(file_path)
            close_wb(wb)

            self.opened_excel_files.pop(file_path)

    def release(self):
        for file_path in self.opened_excel_files:
            wb = self.opened_excel_files[file_path]
            close_wb(wb)

        self.opened_excel_files.clear()
        quit_app(self.excel_app)

    def get_excel_book(self, err: Error,
                       file_path: str,
                       multi_reading: bool = True):
        if file_path in self.opened_excel_files:
            return self.opened_excel_files[file_path]

        if not should_exists(err, file_path):
            return

        if not multi_reading and is_excel_file_opened(file_path):
            err.append("{} is opened, please close it and try again".format(file_path))
            return

        wb = self.excel_app.books.open(file_path)
        self.opened_excel_files[file_path] = wb

        return wb

    def get_excel_sheet(self, err: Error,
                        file_path: str,
                        sheet_name: Optional[str] = None,
                        multi_reading: bool = True):
        wb: xw.Book = self.get_excel_book(err, file_path, multi_reading)
        if err.has_error():
            return

        if is_none_or_empty(sheet_name):
            sht: xw.Sheet = wb.sheets[0]
        else:
            sht = sheet_with_name(err, wb, name=sheet_name)

        return None if err.has_error() else sht

    def create_excel_file(self, err: Error, file_path: str, template_file_path: Optional[str] = None):
        try:
            if template_file_path is None:
                wb = self.excel_app.books.add()
                wb.save(file_path)
                self.opened_excel_files[file_path] = wb
            else:
                copy(template_file_path, file_path)
                self.get_excel_book(err, file_path, False)
        except Exception as e:
            err.append('create new excel file as {} failed with error {}'.format(file_path, e))
